Package: mvMonitoring
Type: Package
Title: Multi-State Adaptive Dynamic Principal Component Analysis for
        Multivariate Process Monitoring
Version: 0.2.4
Date: 2023-11-21
Authors@R: 
  c(
    person(
      "Melissa", "Innerst", email = "innerst@juniata.edu", role = c("aut")
    ),
    person(
      "Gabriel", "Odom", email = "gabriel.odom@fiu.edu", role = c("aut", "cre")
    ), 
    person(
      "Ben", "Barnard", email = "ben_barnard@baylor.edu", role = c("aut")
    ),
    person("Karen", "Kazor", role = c("aut")),
    person(
      "Amanda", "Hering", email = "mandy_hering@baylor.edu", role = c("aut")
    )
  )
Description: Use multi-state splitting to apply Adaptive-Dynamic PCA (ADPCA) to
    data generated from a continuous-time multivariate industrial or natural
    process. Employ PCA-based dimension reduction to extract linear combinations
    of relevant features, reducing computational burdens. For a description of 
    ADPCA, see <doi:10.1007/s00477-016-1246-2>, the 2016 paper from Kazor et al.
    The multi-state application of ADPCA is from a manuscript under current
    revision entitled "Multi-State Multivariate Statistical Process Control" by
    Odom, Newhart, Cath, and Hering, and is  expected to appear in Q1 of 2018.
License: GPL-2
Depends: R (>= 2.10)
Imports: dplyr, lazyeval, plyr, rlang, utils, xts, zoo, robustbase,
        graphics
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.2.3
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
URL: https://github.com/gabrielodom/mvMonitoring
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2023-11-21 17:15:35 UTC; gabrielodom
Author: Melissa Innerst [aut],
  Gabriel Odom [aut, cre],
  Ben Barnard [aut],
  Karen Kazor [aut],
  Amanda Hering [aut]
Maintainer: Gabriel Odom <gabriel.odom@fiu.edu>
Repository: CRAN
Date/Publication: 2023-11-21 17:30:02 UTC
Built: R 4.6.0; ; 2026-04-29 05:30:15 UTC; unix
