# Generated by roxygen2: do not edit by hand

S3method(faultDetect,threshold)
S3method(pca,matrix)
S3method(pca,xts)
S3method(quantile,density)
S3method(threshold,pca)
export(dataStateSwitch)
export(faultDetect)
export(faultFilter)
export(faultSwitch)
export(mspContributionPlot)
export(mspMonitor)
export(mspProcessData)
export(mspSPEPlot)
export(mspSubset)
export(mspT2Plot)
export(mspTrain)
export(mspWarning)
export(pca)
export(processMonitor)
export(processNOCdata)
export(rotate3D)
export(rotateScale3D)
export(threshold)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,hist)
importFrom(graphics,points)
importFrom(lazyeval,lazy_dots)
importFrom(lazyeval,lazy_eval)
importFrom(plyr,ldply)
importFrom(rlang,.data)
importFrom(robustbase,adjbox)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,density)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(utils,head)
importFrom(xts,is.xts)
importFrom(xts,lag.xts)
importFrom(xts,xts)
importFrom(zoo,index)
